#include <map>

#include "ttl_define.h"
#include "tt_util.h"
#include "tt_exception.h"

#include "tt_command_line_analyzer.h"

// -- TtCommandLineOption ------------------------------------------------
TtCommandLineOption::TtCommandLineOption( const std::string& main_option, unsigned int operand_count, bool is_indispensable ) :
main_option_( main_option ),
option_strings_(),
operand_(),
operand_count_( operand_count ),
is_indispensable_( is_indispensable ),
is_specified_( false )
{
  option_strings_.insert( main_option_ );
}

TtCommandLineOption::~TtCommandLineOption()
{
}


const std::string&
TtCommandLineOption::GetMainOption( void ) const
{
  return main_option_;
}


void
TtCommandLineOption::AddOptionString( const std::string& option )
{
  option_strings_.insert( option );
}

const std::set<std::string>&
TtCommandLineOption::GetOptionStrings( void ) const
{
  return option_strings_;
}

void
TtCommandLineOption::SetIndispensable( bool flag )
{
  is_indispensable_ = flag;
}

bool
TtCommandLineOption::IsIndispensable( void ) const
{
  return is_indispensable_;
}

bool
TtCommandLineOption::IsSpecified( void ) const
{
  return is_specified_;
}

const std::vector<std::string>&
TtCommandLineOption::GetOperand( void ) const
{
  return operand_;
}

// -- TtCommandLineOptionBoolean -----------------------------------------
TtCommandLineOptionBoolean::TtCommandLineOptionBoolean( const std::string& option, bool is_indispensable ) :
TtCommandLineOption( option, 0, is_indispensable )
{
}


// -- TtCommandLineOptionSingleOperand -----------------------------------
TtCommandLineOptionSingleOperand::TtCommandLineOptionSingleOperand( const std::string& option, bool is_indispensable ) :
TtCommandLineOption( option, 1, is_indispensable )
{
}

const std::string&
TtCommandLineOptionSingleOperand::GetSingleOperand( void ) const
{
  return this->GetOperand()[0];
}


// -- TtCommandLineOptionAnalyzer ----------------------------------------
TtCommandLineOptionAnalyzer::TtCommandLineOptionAnalyzer() :
options_()
{
}

TtCommandLineOptionAnalyzer::~TtCommandLineOptionAnalyzer()
{
}

void
TtCommandLineOptionAnalyzer::Regist( TtCommandLineOption& option )
{
  options_.insert( &option );
}

std::vector<std::string>
TtCommandLineOptionAnalyzer::AnalyzeForConsole( unsigned int argc, const char* const* argv )
{
  std::vector<std::string> args;
  for ( unsigned int i = 0; i < argc; ++i ) {
    args.push_back( argv[i] );
  }
  return this->Analyze( args );
}

std::vector<std::string>
TtCommandLineOptionAnalyzer::AnalyzeForWindows( const std::string& args_string )
{
  std::vector<std::string> args;
  std::string tmp;
  enum {
    DEFAULT,
    IN_DOUBLE_QUOTE,
    SEPARATOR,
  } status = DEFAULT;

  for ( unsigned int i = 0; i < args_string.size(); ++i ) {
    const char c = args_string[i];
    switch ( status ) {
    case DEFAULT:
      switch ( c ) {
      case ' ':
      case '\t':
        status = SEPARATOR;
        args.push_back( tmp );
        tmp.clear();
        break;

      case '\"':
        status = IN_DOUBLE_QUOTE;
        break;

      case '\\':
        if ( i + 1 < args_string.size() && args_string[i + 1] == '\"' ) {
          tmp.append( 1, '\"' );
          i += 2;
        }
        break;

      default:
        tmp.append( 1, c );
        break;
      }
      break;

    case IN_DOUBLE_QUOTE:
      switch ( c ) {
      case '\"':
        status = DEFAULT;
        break;

      case '\\':
        if ( i + 1 < args_string.size() && args_string[i + 1] == '\"' ) {
          tmp.append( 1, '\"' );
          i += 2;
        }
        break;

      default:
        tmp.append( 1, c );
        break;
      }
      break;

    case SEPARATOR:
      switch ( c ) {
      case ' ':
      case '\t':
        break;

      default:
        --i;
        status = DEFAULT;
        break;
      }
      break;
    }
  }

  if ( NOT( tmp.empty() ) ) {
    args.push_back( tmp );
  }
  return this->Analyze( args );
}


std::vector<std::string>
TtCommandLineOptionAnalyzer::Analyze( const std::vector<std::string> args )
{
  std::map<std::string, TtCommandLineOption*> table;
  for ( std::set<TtCommandLineOption*>::const_iterator option = options_.begin(); option != options_.end(); ++option ) {
    const std::set<std::string>& option_strings = (*option)->GetOptionStrings();
    for ( std::set<std::string>::const_iterator key = option_strings.begin(); key != option_strings.end(); ++key ) {
      if ( table.find( *key ) != table.end() ) {
        throw TtCommandLineOptionDuplicateDefinitionException( *key );
      }
      table[*key] = *option;
    }
  }

  std::vector<std::string> rest;
  for ( std::vector<std::string>::const_iterator it = args.begin(); it != args.end(); ++it ) {
    if ( (*it)[0] == '-' ) {
      std::string option_string = it->substr( 1 );
      if ( table.find( option_string ) == table.end() ) {
        throw TtCommandLineOptionUnknownOptionSpecifiedException( (*it) );
      }
      TtCommandLineOption& option = *table[it->substr( 1 )];

      if ( option.is_specified_ ) {
        throw TtCommandLineOptionDuplicatedSpecifyException( option_string );
      }
      option.is_specified_ = true;

      for ( unsigned int i = 0; i < option.operand_count_; ++i ) {
        ++it;
        if ( it == args.end() || (*it)[0] == '-' ) {
          throw TtCommandLineOptionOperandCountException( option_string );
        }
        option.operand_.push_back( *it );
      }
    }
    else {
      rest.push_back( *it );
    }
  }

  for ( std::map<std::string, TtCommandLineOption*>::iterator it = table.begin(); it != table.end(); ++it ) {
    if ( it->second->is_indispensable_ && NOT( it->second->is_specified_ ) ) {
      throw TtCommandLineOptionIndispensableIsNoneException( it->second->GetMainOption() );
    }
  }
  return rest;
}
