#include "ttl_define.h"
#include "tt_exception.h"
#include "tt_file_dialog.h"
#include "tt_window_controls.h"

// -- TtFileDialogFilter -------------------------------------------------
TtFileDialogFilter::TtFileDialogFilter( void ) :
description_(),
extensions_()
{
}

TtFileDialogFilter::TtFileDialogFilter( const std::string& description, const std::string& extension ) :
description_( description ),
extensions_()
{
  extensions_.push_back( extension );
}

const std::string&
TtFileDialogFilter::GetDescription( void ) const
{
  return description_;
}

void
TtFileDialogFilter::SetDescription( const std::string& description )
{
  description_ = description;
}

std::vector<std::string>&
TtFileDialogFilter::GetExtensions( void )
{
  return extensions_;
}

// -- TtFileDialog -------------------------------------------------------
TtFileDialog::TtFileDialog( unsigned int buffer_size ) :
check_file_exists_( false ),
check_path_exists_( true ),
default_extension_( "" ),
dereference_links_( true ),
file_name_( "" ),
filter_index_( 1 ),
initial_directory_( "" ),
restore_directory_( false ),
show_help_( false ),
show_place_bar_( true ),
title_( "" ),
file_names_(),
filters_(),
buffer_( buffer_size )
{
}


bool
TtFileDialog::GetCheckFileExists( void ) const
{
  return check_file_exists_;
}

void
TtFileDialog::SetCheckFileExists( bool flag )
{
  check_file_exists_ = flag;
}

bool
TtFileDialog::GetCheckPathExists( void ) const
{
  return check_path_exists_;
}

void
TtFileDialog::SetCheckPathExists( bool flag )
{
  check_path_exists_ = flag;
}

const std::string&
TtFileDialog::GetDefaultExtension( void ) const
{
  return default_extension_;
}

void
TtFileDialog::SetDefaultExtension( const std::string& extension )
{
  default_extension_ = extension;
}

bool
TtFileDialog::GetDereferenceLinks( void ) const
{
  return dereference_links_;
}

void
TtFileDialog::SetDereferenceLinks( bool flag )
{
  dereference_links_ = flag;
}

const std::string&
TtFileDialog::GetFileName( void ) const
{
  return file_name_;
}

void
TtFileDialog::SetFileName( const std::string& file_name )
{
  file_name_ = file_name;
}

int
TtFileDialog::GetFilterIndex( void ) const
{
  return filter_index_;
}

void
TtFileDialog::SetFilterIndex( int index )
{
  filter_index_ = index;
}

const std::string&
TtFileDialog::GetInitialDirectory( void ) const
{
  return initial_directory_;
}

void
TtFileDialog::SetInitialDirectory( const std::string& directory )
{
  initial_directory_ = directory;
}

bool
TtFileDialog::GetRestoreDirectory( void ) const
{
  return restore_directory_;
}

void
TtFileDialog::SetRestoreDirectory( bool flag )
{
  restore_directory_ = flag;
}

bool
TtFileDialog::GetShowHelp( void ) const
{
  return show_help_;
}

void
TtFileDialog::SetShowHelp( bool flag )
{
  show_help_ = flag;
}

const std::string&
TtFileDialog::GetTitle( void ) const
{
  return title_;
}

void
TtFileDialog::SetTitle( const std::string& title )
{
  title_ = title;
}

bool
TtFileDialog::GetShowPlaceBar( void ) const
{
  return show_place_bar_;
}

void
TtFileDialog::SetShowPlaceBar( bool flag )
{
  show_place_bar_ = flag;
}

const std::vector<std::string>&
TtFileDialog::GetFileNames( void ) const
{
  return file_names_;
}

std::vector<TtFileDialogFilter>&
TtFileDialog::GetFilters( void )
{
  return filters_;
}

void
TtFileDialog::Reset( void )
{
  check_file_exists_ = false;
  check_path_exists_ = true;
  default_extension_ = "";
  dereference_links_ = true;
  file_name_         = "";
  filter_index_      = 1;
  initial_directory_ = "";
  restore_directory_ = false;
  show_help_         = false;
  title_             = "";
  file_names_.clear();
  filters_.clear();
}

bool
TtFileDialog::ShowDialog( void )
{
  TtStatic dummy;
  return this->ShowDialog( dummy );
}


TtTemporalyString&
TtFileDialog::GetBuffer( void )
{
  return buffer_;
}

std::string
TtFileDialog::MakeFilterString( void )
{
  std::string str;
  for ( std::vector<TtFileDialogFilter>::iterator filter = filters_.begin(); filter != filters_.end(); ++filter ) {
    str.append( filter->GetDescription() );
    str.append( 1, '\0' );
    for ( std::vector<std::string>::iterator ext = filter->GetExtensions().begin(); ext != filter->GetExtensions().end(); ++ext ) {
      str.append( "*." );
      str.append( *ext );
      str.append( ";" );
    }
    str.erase( str.size() - 1 );
    str.append( 1, '\0' );
  }
  str.append( 1, '\0' );
  return str;
}

void
TtFileDialog::MakeStructure( OPENFILENAME& ofn )
{
  memset( &ofn, 0, sizeof( OPENFILENAME ) );

  ofn.lStructSize     = sizeof( OPENFILENAME );

  file_name_.copy( buffer_.GetPtr(), buffer_.GetSize() );
  buffer_.GetPtr()[file_name_.size() < buffer_.GetSize() ? file_name_.size() : buffer_.GetSize()] = '\0';
  ofn.lpstrFile = buffer_.GetPtr();
  ofn.nMaxFile = buffer_.GetSize();

  if ( NOT( initial_directory_.empty() ) ) {
    ofn.lpstrInitialDir = initial_directory_.c_str();
  }

  if ( NOT( title_.empty() ) ) {
    ofn.lpstrTitle = title_.c_str();
  }

  if ( NOT( default_extension_.empty() ) ) {
    ofn.lpstrDefExt = default_extension_.c_str();
  }

  ofn.Flags |= this->GetCheckFileExists()  ? OFN_FILEMUSTEXIST : 0;
  ofn.Flags |= this->GetCheckPathExists()  ? OFN_PATHMUSTEXIST : 0;
  ofn.Flags |= this->GetDereferenceLinks() ? 0 : OFN_NODEREFERENCELINKS;
  ofn.Flags |= this->GetRestoreDirectory() ? OFN_NOCHANGEDIR : 0;
  ofn.Flags |= this->GetShowHelp()         ? OFN_SHOWHELP : 0;
  ofn.Flags |= OFN_LONGNAMES;
  ofn.Flags |= OFN_EXPLORER;

  ofn.FlagsEx |= this->GetShowPlaceBar() ? 0 : OFN_EX_NOPLACESBAR;
}

void
TtFileDialog::SetFileNames( void )
{
  file_name_ = buffer_.GetPtr();
  file_names_.clear();
  const char* p = buffer_.GetPtr();
  for (;;) {
    file_names_.push_back( std::string( p ) );
    while ( *p != '\0' ) {
      ++p;
    }
    ++p;
    if ( *p == '\0' ) {
      break;
    }
  }
}


// -- TtOpenFileDialog ---------------------------------------------------
TtOpenFileDialog::TtOpenFileDialog( void ) :
TtFileDialog( MAX_PATH * 128 ),
multiselect_( false ),
read_only_checked_( false ),
show_read_only_( false )
{
  this->SetCheckFileExists( true );
}

TtOpenFileDialog::TtOpenFileDialog( unsigned int buffer_size ) :
TtFileDialog( buffer_size ),
multiselect_( false ),
read_only_checked_( false ),
show_read_only_( false )
{
  this->SetCheckFileExists( true );
}


bool
TtOpenFileDialog::GetMultiselect( void ) const
{
  return multiselect_;
}

void
TtOpenFileDialog::SetMultiselect( bool flag )
{
  multiselect_ = flag;
}

bool
TtOpenFileDialog::GetReadOnlyChecked( void ) const
{
  return read_only_checked_;
}

void
TtOpenFileDialog::SetReadOnlyChecked( bool flag )
{
  read_only_checked_ = flag;
}

bool
TtOpenFileDialog::GetShowReadOnly( void ) const
{
  return show_read_only_;
}

void
TtOpenFileDialog::SetShowReadOnly( bool flag )
{
  show_read_only_ = flag;
}

void
TtOpenFileDialog::Reset( void )
{
  this->TtFileDialog::Reset();
  multiselect_       = false;
  read_only_checked_ = false;
  show_read_only_    = false;
  this->SetCheckFileExists( true );
}

bool
TtOpenFileDialog::ShowDialog( TtWindow& parent )
{
  OPENFILENAME ofn;
  this->MakeStructure( ofn );

  ofn.hwndOwner = parent.GetHandle();

  std::string filter_str = this->MakeFilterString();
  TtTemporalyString filter_buf( filter_str.size() + 1 );
  filter_str.copy( filter_buf.GetPtr(), filter_str.size() );
  filter_buf.GetPtr()[filter_buf.GetSize() - 1] = '\0';

  if ( NOT( this->GetFilters().empty() ) ) {
    ofn.lpstrFilter  = filter_buf.GetPtr();
    ofn.nFilterIndex = this->GetFilterIndex();
  }

  ofn.Flags |= this->GetMultiselect()     ? OFN_ALLOWMULTISELECT : 0;
  ofn.Flags |= this->GetReadOnlyChecked() ? OFN_READONLY : 0;
  ofn.Flags |= this->GetShowReadOnly()    ? 0 : OFN_HIDEREADONLY;

  bool ret = GetOpenFileName( &ofn );

  this->SetFileNames();
  return ret;
}

// -- TtSaveFileDialog ---------------------------------------------------
TtSaveFileDialog::TtSaveFileDialog( void ) :
TtFileDialog( MAX_PATH ),
create_prompt_( false ),
overwite_prompt_( true )
{
}


bool
TtSaveFileDialog::GetCreatePrompt( void ) const
{
  return create_prompt_;
}

void
TtSaveFileDialog::SetCreatePrompt( bool flag )
{
  create_prompt_ = flag;
}

bool
TtSaveFileDialog::GetOverwritePrompt( void ) const
{
  return overwite_prompt_;
}

void
TtSaveFileDialog::SetOverwritePrompt( bool flag )
{
  overwite_prompt_ = flag;
}

void
TtSaveFileDialog::Reset( void )
{
  this->TtFileDialog::Reset();
  create_prompt_   = false;
  overwite_prompt_ = true;
}

bool
TtSaveFileDialog::ShowDialog( TtWindow& parent )
{
  OPENFILENAME ofn;
  this->MakeStructure( ofn );

  ofn.hwndOwner = parent.GetHandle();

  std::string filter_str = this->MakeFilterString();
  TtTemporalyString filter_buf( filter_str.size() + 1 );
  filter_str.copy( filter_buf.GetPtr(), filter_str.size() );
  filter_buf.GetPtr()[filter_buf.GetSize() - 1] = '\0';

  if ( NOT( this->GetFilters().empty() ) ) {
    ofn.lpstrFilter  = filter_buf.GetPtr();
    ofn.nFilterIndex = this->GetFilterIndex();
  }

  ofn.Flags |= this->GetCreatePrompt()    ? OFN_CREATEPROMPT : 0;
  ofn.Flags |= this->GetOverwritePrompt() ? OFN_OVERWRITEPROMPT : 0;

  bool ret = GetSaveFileName( &ofn );

  this->SetFileNames();
  return ret;
}
