#include <windows.h>

#include "ttl_define.h"
#include "tt_exception.h"
#include "tt_window.h"
#include "tt_font.h"
#include "tt_temporaly_array.h"

// -----------------------------------------------------------------------
const int
TtWindow::MSG_DEFAULT = TT_WINDOW_MSG_DEFAULT;

std::map<HWND, TtWindow*>
TtWindow::TT_WINDOW_TABLE;


void
TtWindow::InitializeCommonControls( void )
{
  InitCommonControls();
  CoInitialize( NULL );
}

LRESULT CALLBACK
TtWindow::WindowProcedureForTTL( HWND handle, UINT msg, WPARAM w_param, LPARAM l_param )
{
  TtWindow* window = TtWindow::FindWindow( handle );
  return ( window != NULL ) ?
    window->WindowProcedure( msg, w_param, l_param ) : DefWindowProc( handle, msg, w_param, l_param );
}

TtWindow*
TtWindow::FindWindow( HWND handle )
{
  std::map<HWND, TtWindow*>::iterator it = TT_WINDOW_TABLE.find( handle );
  return ( it != TT_WINDOW_TABLE.end() ) ? it->second : NULL;
}

std::map<HWND, TtWindow*>&
TtWindow::GetTtWindowTable( void )
{
  return TT_WINDOW_TABLE;
}


// -----------------------------------------------------------------------
bool
TtWindow::Created( void )
{
  return true;
}

LRESULT
TtWindow::WindowProcedure( UINT msg, WPARAM w_param, LPARAM l_param )
{
  return ( window_procedure_super_ == NULL ) ?
    this->DefaultWindowProcedure( msg, w_param, l_param )
      : CallWindowProc( window_procedure_super_, handle_, msg, w_param, l_param );
}

LRESULT
TtWindow::WMHorizontalScrollReceived( int code, int position, HWND handle )
{
  NOT_USE_VAR( code );
  NOT_USE_VAR( position );
  NOT_USE_VAR( handle );
  return TtWindow::MSG_DEFAULT;
}

LRESULT
TtWindow::WMVerticalScrollReceived( int code, int position, HWND handle )
{
  NOT_USE_VAR( code );
  NOT_USE_VAR( position );
  NOT_USE_VAR( handle );
  return TtWindow::MSG_DEFAULT;
}

LRESULT
TtWindow::DefaultWindowProcedure( UINT msg, WPARAM w_param, LPARAM l_param )
{
  return DefWindowProc( handle_, msg, w_param, l_param );
}


TtWindow::TtWindow( void ) :
parent_( NULL ),
handle_( NULL ),
id_( 0 ),
instance_handle_( NULL ),
window_procedure_super_( NULL )
{
}

TtWindow::~TtWindow()
{
  if ( handle_ != NULL && TT_WINDOW_TABLE.find( handle_ ) != TT_WINDOW_TABLE.end() ) {
    TT_WINDOW_TABLE.erase( handle_ );
  }
}


// -- CreateParameter ----------------------------------------------------
TtWindow::CreateParameter::CreateParameter( void ) :
parent_( NULL ),
id_( 0 ),
instance_handle_( GetModuleHandle( NULL ) ),
lp_param_( NULL ),
is_child_( true ),
group_start_( false )
{
}

TtWindow::CreateParameter::CreateParameter( TtWindow* parent ) :
parent_( parent ),
id_( 0 ),
instance_handle_( parent->GetInstanceHandle() ),
lp_param_( NULL ),
is_child_( true ),
group_start_( false )
{
}

TtWindow::CreateParameter::CreateParameter( TtWindow* parent, int id ) :
parent_( parent ),
id_( id ),
instance_handle_( parent->GetInstanceHandle() ),
lp_param_( NULL ),
is_child_( true ),
group_start_( false )
{
}

TtWindow::CreateParameter::CreateParameter( const CreateParameter& other ) :
parent_( other.parent_ ),
id_( other.id_ ),
instance_handle_( other.instance_handle_ ),
lp_param_( other.lp_param_ ),
is_child_( other.is_child_ ),
group_start_( other.group_start_ )
{
}

TtWindow::CreateParameter
TtWindow::CreateParameter::AlterParent( TtWindow* parent ) const
{
  CreateParameter tmp( *this );
  tmp.parent_ = parent;
  return tmp;
}

TtWindow::CreateParameter
TtWindow::CreateParameter::AlterId( int id ) const
{
  CreateParameter tmp( *this );
  tmp.id_ = id;
  return tmp;
}

TtWindow::CreateParameter
TtWindow::CreateParameter::AlterInstanceHandle( HINSTANCE instance_handle ) const
{
  CreateParameter tmp( *this );
  tmp.instance_handle_ = instance_handle;
  return tmp;
}

TtWindow::CreateParameter
TtWindow::CreateParameter::AlterLPParam( LPVOID lp_param ) const
{
  CreateParameter tmp( *this );
  tmp.lp_param_ = lp_param;
  return tmp;
}

TtWindow::CreateParameter
TtWindow::CreateParameter::AlterIsChild( bool is_child ) const
{
  CreateParameter tmp( *this );
  tmp.is_child_ = is_child;
  return tmp;
}

TtWindow::CreateParameter
TtWindow::CreateParameter::AlterGroupStart( bool group_start ) const
{
  CreateParameter tmp( *this );
  tmp.group_start_ = group_start;
  return tmp;
}



void
TtWindow::Create( void )
{
  this->Create( CreateParameter() );
}

void
TtWindow::Create( CreateParameter parameter )
{
  this->CreatePrimitive( parameter );
}


void
TtWindow::CreatePrimitive( CreateParameter& parameter )
{
  parent_ = parameter.parent_;
  instance_handle_ = parameter.instance_handle_;
  id_ = parameter.id_;

  DWORD style = ( this->GetStyle() |
                  ((parameter.parent_ != NULL && parameter.is_child_) ? WS_CHILD : 0) |
                  (parameter.group_start_ ? WS_GROUP : 0) );

  handle_ = CreateWindowEx(
    this->GetExtendedStyle(),
    this->GetClassName(),
    NULL,
    style,
    CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
    parameter.parent_ ? parameter.parent_->GetHandle() : NULL,
    reinterpret_cast<HMENU>( parameter.id_ ),
    instance_handle_,
    parameter.lp_param_ );
  if ( handle_ == NULL ) {
    throw TT_WIN_SYSTEM_CALL_EXCEPTION( FUNC_NAME_OF( CreateWindowEx ) );
  }

  this->SetFont( TtFont::DEFAULT );
  TT_WINDOW_TABLE.insert( std::pair<HWND, TtWindow*>( handle_, this ) );

  if ( NOT( this->Created() ) ) {
    ::DestroyWindow( handle_ );
    handle_ = NULL;
  }
}


int
TtWindow::Close( void )
{
  return this->SendMessage( WM_CLOSE );
}


int
TtWindow::SendMessage( UINT msg, WPARAM w_param, LPARAM l_param )
{
  return ::SendMessage( handle_, msg, w_param, l_param );
}

void
TtWindow::PostMessage( UINT msg, WPARAM w_param, LPARAM l_param )
{
  if ( ::PostMessage( handle_, msg, w_param, l_param ) == 0 ) {
    throw TT_WIN_SYSTEM_CALL_EXCEPTION( FUNC_NAME_OF( ::PostMessage ) );
  }
}


void
TtWindow::SetWindowLong( int index, LONG value )
{
  SetLastError( 0 );
  ::SetWindowLong( handle_, index, value );
  if ( GetLastError() != 0 ) {
    throw TT_WIN_SYSTEM_CALL_EXCEPTION( FUNC_NAME_OF( ::SetWindowLong ) );
  }
}

LONG
TtWindow::GetWindowLong( int index )
{
  SetLastError( 0 );
  LONG tmp = ::GetWindowLong( handle_, index );
  if ( GetLastError() != 0 ) {
    throw TT_WIN_SYSTEM_CALL_EXCEPTION( FUNC_NAME_OF( ::GetWindowLong ) );
  }
  return tmp;
}


void
TtWindow::OverrideWindowProcedure( void )
{
  window_procedure_super_ = reinterpret_cast<WNDPROC>( this->GetWindowLong( GWL_WNDPROC ) );
  this->SetWindowLong( GWL_WNDPROC, reinterpret_cast<LONG>( TtWindow::WindowProcedureForTTL ) );
}


TtWindow*
TtWindow::GetParent( void )
{
  return parent_;
}

HWND
TtWindow::GetHandle( void )
{
  return handle_;
}

int
TtWindow::GetID( void )
{
  return id_;
}

HINSTANCE
TtWindow::GetInstanceHandle( void )
{
  return instance_handle_;
}


RECT
TtWindow::GetRectangle( void )
{
  RECT rect;
  if ( GetWindowRect( handle_, &rect ) == 0 ) {
    throw TT_WIN_SYSTEM_CALL_EXCEPTION( FUNC_NAME_OF( GetWindowRect ) );
  }
  return rect;
}

int
TtWindow::GetHorizontalPosition( void )
{
  return this->GetRectangle().left;
}

int
TtWindow::GetVerticalPosition( void )
{
  return this->GetRectangle().top;
}

int
TtWindow::GetWidth( void )
{
  RECT rect = this->GetRectangle();
  return rect.right - rect.left;
}

int
TtWindow::GetHeight( void )
{
  RECT rect = this->GetRectangle();
  return rect.bottom - rect.top;
}

void
TtWindow::Show( void )
{
  this->SetShowState( SW_SHOW );
}

void
TtWindow::Hide( void )
{
  this->SetShowState( SW_HIDE );
}

void
TtWindow::Minimize( void )
{
  this->SetShowState( SW_MINIMIZE );
}

void
TtWindow::Maximize( void )
{
  this->SetShowState( SW_MAXIMIZE );
}

void
TtWindow::Restore()
{
  this->SetShowState( SW_RESTORE );
}

void
TtWindow::SetShowState( int state )
{
  ShowWindow( handle_, state );
}

bool
TtWindow::GetVisible( void )
{
  return IsWindowVisible( handle_ );
}


bool
TtWindow::GetEnabled( void )
{
  return IsWindowEnabled( handle_ );
}

void
TtWindow::SetEnabled( bool flag )
{
  EnableWindow( handle_, flag );
}


void
TtWindow::SetActive( void )
{
  SetActiveWindow( handle_ );
}


void
TtWindow::SetFocus( void )
{
  ::SetFocus( handle_ );
}


void
TtWindow::SetPosition( int x, int y )
{
  RECT rect = this->GetRectangle();
  this->Move( x, y, rect.right - rect.left, rect.bottom - rect.top, TRUE );
}

void
TtWindow::SetSize( int width, int height )
{
  RECT rect = this->GetRectangle();
  this->Move( rect.left, rect.top, width, height, TRUE );
}

void
TtWindow::SetPositionSize( int x, int y, int width, int height )
{
  this->Move( x, y, width, height, TRUE );
}

void
TtWindow::SetRectangle( RECT rect )
{
  this->Move( rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top, TRUE );
}


void
TtWindow::Move( int x, int y, int width, int height, int repaint )
{
  if ( MoveWindow( handle_, x, y, width, height, repaint ) == 0 ) {
    throw TT_WIN_SYSTEM_CALL_EXCEPTION( FUNC_NAME_OF( MoveWindow ) );
  }
}

void
TtWindow::SetCenterRelativeToParent( void )
{
  RECT parent_rect = parent_->GetRectangle();
  RECT this_rect = this->GetRectangle();

  int x = (parent_rect.right + parent_rect.left) / 2 - (this_rect.right - this_rect.left) / 2;
  if ( x < 0 ) {
    x = 0;
  }
  else if ( x + this_rect.right - this_rect.left > TtExtraordinarilyWindow::Desktop.GetRectangle().right ) {
    x = TtExtraordinarilyWindow::Desktop.GetRectangle().right - this->GetWidth();
  }

  int y = (parent_rect.bottom + parent_rect.top) / 2 - (this_rect.bottom - this_rect.top) / 2;
  if ( y < 0 ) {
    y = 0;
  }
  else if ( y + this_rect.bottom - this_rect.top > TtExtraordinarilyWindow::Desktop.GetRectangle().bottom ) {
    y = TtExtraordinarilyWindow::Desktop.GetRectangle().bottom - this->GetHeight();
  }

  this->SetPosition( x, y );
}


void
TtWindow::SetBottom( void )
{
  this->SetUnder( HWND_BOTTOM );
}

void
TtWindow::SetTop( void )
{
  this->SetUnder( HWND_TOP );
}

void
TtWindow::SetUnder( TtWindow& window )
{
  this->SetUnder( window.handle_ );
}

void
TtWindow::SetUnder( HWND window )
{
  if ( SetWindowPos( handle_, window, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE ) == 0 ) {
    throw TT_WIN_SYSTEM_CALL_EXCEPTION( FUNC_NAME_OF( MoveWindow ) );
  }
}

void
TtWindow::SetTopMost( bool flag )
{
  if ( SetWindowPos( handle_, flag ? HWND_TOPMOST : HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE ) == 0 ) {
    throw TT_WIN_SYSTEM_CALL_EXCEPTION( FUNC_NAME_OF( MoveWindow ) );
  }
}


void
TtWindow::SetFont( TtFont& font )
{
  this->SendMessage( WM_SETFONT, reinterpret_cast<WPARAM>( font.GetHandle() ), true );
}

void
TtWindow::SetIcon( HICON icon, bool is_big )
{
  this->SendMessage( WM_SETICON, is_big ? ICON_BIG : ICON_SMALL, reinterpret_cast<LPARAM>( icon ) );
}



int
TtWindow::SetForground( void )
{
  return SetForegroundWindow( handle_ );
}


int
TtWindow::GetTextLength( void )
{
  return ::GetWindowTextLength( handle_ );
}

std::string
TtWindow::GetText( void )
{
  int length = this->GetTextLength();
  if ( length == 0 ) {
    return std::string();
  }
  TtTemporalyString buf( length + 1 );
  if ( ::GetWindowText( handle_, buf.GetPtr(), buf.GetSize() ) == 0 ) {
    throw TT_WIN_SYSTEM_CALL_EXCEPTION( FUNC_NAME_OF( GetWindowText ) );
  }
  return std::string( buf.GetPtr() );
}

void
TtWindow::SetText( const std::string& str )
{
  if ( SetWindowText( handle_, str.c_str() ) == 0 ) {
    throw TT_WIN_SYSTEM_CALL_EXCEPTION( FUNC_NAME_OF( SetWindowText ) );
  }
}



// -- TtExtraordinarilyWindow --------------------------------------------
TtExtraordinarilyWindow
TtExtraordinarilyWindow::Desktop( GetDesktopWindow() );

TtExtraordinarilyWindow::TtExtraordinarilyWindow( HWND handle ) :
TtWindow()
{
  handle_ = handle;
}

TtExtraordinarilyWindow::TtExtraordinarilyWindow( HWND handle, TtWindow& parent ) :
TtWindow()
{
  handle_ = handle;
  parent_ = &parent;
}

DWORD
TtExtraordinarilyWindow::GetStyle( void )
{
  return 0;
}

DWORD
TtExtraordinarilyWindow::GetExtendedStyle( void )
{
  return 0;
}

PCTSTR
TtExtraordinarilyWindow::GetClassName( void )
{
  return 0;
}
